from inspect import signature
from pygame import font


def find_num_params(func: callable) -> int:
    """Return the number of arguments a function accepts."""
    sig = signature(func)
    params = sig.parameters

    return len(params)


def call_func(func: callable, *args) -> None:
    """
    Call a function - if it is not None - and pass the appropriate arguments in to it.
    
    An exception is raised if the function accepts an invalid number of arguments.
    """

    if func == None:
        return
    
    num_params = find_num_params(func)

    if num_params == 0:
        func()
    elif num_params == 1:
        func(args[0])
    elif num_params == 2:
        func(args[0], args[1])
    else:
        raise Exception(f"Invalid number of parameters for {func}. {func} should accept 0, 1 or 2 arguments.")
    

def update_font_attrs(obj: object, text: str, font_colour: tuple[int], font_name: str, font_size: int) -> None:
    """Change the text, font colour and font attributes of an object."""
    obj.text = text
    obj.font_colour = font_colour

    obj.font = font.Font(font_name, font_size)


from pygame_ui_toolkit.elements import slider
from pygame_ui_toolkit import pygame


stored_variables = {}


def assign_variables(slider_object: slider.HorizontalSlider | slider.VerticalSlider, surface: pygame.Surface, font_name: str | None, font_size: int, font_colour: tuple[int], antialias: bool, x: int, y: int, int_only: bool) -> None:
    """Assign variables that need to be used later to a global dict with the slider object as the key."""

    current_variables = {
        "surface" : surface,

        "font" : pygame.font.Font(font_name, font_size),
        "font_colour" : font_colour,
        "antialias" : antialias,

        "text_x" : x,
        "text_y" : y,

        "use_int" : int_only
    }
    
    stored_variables[slider_object] = current_variables

    on_value_changed(slider_object.value, slider_object)


def on_value_changed(value: float, slider_object: slider.HorizontalSlider | slider.VerticalSlider) -> None:
    """Update the text to display the new value of the slider."""

    variables = stored_variables[slider_object]
    
    if variables["use_int"]:
        value = str(int(value))
    else:
        value = f"{value :.1f}"

    text_surface = variables["font"].render(value, variables["antialias"], variables["font_colour"])

    text_rect = text_surface.get_rect()
    text_rect.center = (variables["text_x"], variables["text_y"])

    variables["text_surface"] = text_surface
    variables["text_rect"] = text_rect

    variables["surface"].blit(text_surface, text_rect)


def get_slider(is_horizontal: bool, surface: pygame.Surface, length: int, width: int, x: int, y: int, min_value: float, max_value: float, start_value: float, slider_colour: tuple[int], slider_button: object | None):
    """Return a slider object."""
    if is_horizontal:
        slider_object = slider.HorizontalSlider(surface, length, width, x, y, min_value, max_value, start_value, slider_colour, on_value_changed, slider_button)
    else:
        slider_object = slider.VerticalSlider(surface, length, width, x, y, min_value, max_value, start_value, slider_colour, on_value_changed, slider_button)

    return slider_object


def blit_slider_text(slider_object: slider.HorizontalSlider | slider.VerticalSlider) -> None:
    """
    Draw the slider text to the screen.

    This should be called once per frame alongside the update() method for the slider object.
    """
    
    variables = stored_variables[slider_object]

    variables["surface"].blit(variables["text_surface"], variables["text_rect"])


def create_slider_text(surface: pygame.Surface, slider_x: int, slider_y: int, text_x: int, text_y: int, slider_length: int, slider_width: int, min_value: float, max_value: float, start_value: float, slider_colour: tuple[int], font_name: str | None, font_size: int, font_colour: tuple[int], antialias: bool = False, horizontal_slider: bool = True, int_only: bool = False, slider_button: object | None = None) -> slider.HorizontalSlider | slider.VerticalSlider:
    """
    Return a slider object with text that displays its value.

    Notes
    -----
    - if horizontal_slider is set to True, a horizontal slider will be created. Otherwise, a vertical slider will.
    - if slider_button is set to None, sliders will be created with circular buttons. Otherwise, the provided button_object will be used.
    - if int_only is set to True, the text will only display integer values of the slider. Otherwise, the values will be displayed to 1 decimal place.
    - only the slider object is returned, not the text. To access the text, use stored_variables[slider object]["text_surface"] and stored_variables[slider object]["text_surface"].
    - ensure blit_slider_text and slider.update() are called once per frame.
    """
    
    slider_object = get_slider(horizontal_slider, surface, slider_length, slider_width, slider_x, slider_y, min_value, max_value, start_value, slider_colour, slider_button)
    
    assign_variables(slider_object, surface, font_name, font_size, font_colour, antialias, text_x, text_y, int_only)
    
    return slider_object


from pygame_ui_toolkit.elements import button
from pygame_ui_toolkit import pygame


stored_variables = {}


def assign_variables(button_object: button.Button, normal_colour: tuple[int], hover_colour: tuple[int], click_colour: tuple[int], on_normal: callable, on_hover: callable, on_click: callable) -> None:
    """Assign variables that need to be used later to a global dict with the button object as the key."""
    
    current_variables = {
        "normal_colour" : normal_colour,
        "hover_colour" : hover_colour,
        "click_colour" : click_colour,

        "on_normal" : on_normal,
        "on_hover" : on_hover,
        "on_click" : on_click
    }

    stored_variables[button_object] = current_variables


def update_button_colour(button_object: button.Button, event_type: str) -> None:
    """Change the background colour of the button."""

    variables = stored_variables[button_object]

    button_object.call_func(variables[f"on_{event_type}"])

    button_object.background_colour = variables[f"{event_type}_colour"]


def on_click_func(button_object: button.Button) -> None:
    update_button_colour(button_object, "click")


def on_hover_func(button_object: button.Button) -> None:
    update_button_colour(button_object, "hover")


def on_normal_func(button_object: button.Button) -> None:
    update_button_colour(button_object, "normal")


def create_button(normal_colour: tuple[int], hover_colour: tuple[int], click_colour: tuple[int], surface: pygame.Surface, x: int, y: int, width: int, height: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1) -> button.RectButton:
    """
    Return a new RectButton that will automatically change colour when clicked or hovered.
    
    Ensure to call the button's update() method each frame.
    """

    button_object = button.RectButton(surface, x, y, normal_colour, width, height, on_click_func, on_hover_func, on_normal_func, corner_radius, False)
    
    assign_variables(button_object, normal_colour, hover_colour, click_colour, on_normal, on_hover, on_click)

    return button_object


def change_existing_button(button_object: button.PolygonButton, normal_colour: tuple[int], hover_colour: tuple[int], click_colour: tuple[int]) -> None:
    """
    Update an existing button to make it automatically change colour when clicked or hovered.

    Any button type (other than the base Button class) is supported.

    Note: this will change the click_once attribute to False.
    """

    on_click = button_object.on_click
    on_hover = button_object.on_hover
    on_normal = button_object.on_normal
    
    button_object.on_click = on_click_func
    button_object.on_hover = on_hover_func
    button_object.on_normal = on_normal_func

    button_object.click_once = False

    assign_variables(button_object, normal_colour, hover_colour, click_colour, on_normal, on_hover, on_click)


from pygame_ui_toolkit.elements import toggle
from pygame_ui_toolkit.elements import button
from pygame_ui_toolkit import utils
from pygame_ui_toolkit import pygame


class Option:
    """
    An individual option for the dropdown menu.

    Attributes
    ----------
    parent_dropdown : object
        the dropdown menu object that this option is a part of
    text_wrapper : button.TextWrapper
        the button text wrapper oboject that provides the button functionality, like on click events
    normal_button_on_click : callable
        the text wrapper button's on_click function before it is changed by the Option class
    active : bool
        whether or not this option should be displayed and updated

    Methods
    -------
    setup_button()
        prepare the text_wrapper attribute
    on_button_click()
        call normal_button_on_click and select this option
    update()
        update the object
    """

    def __init__(self, parent_dropdown: object, text_wrapper: button.TextWrapper, start_active: bool = False) -> None:
        """Construct the necessary attributes for the Option object."""
        self.parent_dropdown = parent_dropdown
        
        self.text_wrapper = text_wrapper
        self.normal_button_on_click = text_wrapper.button_object.on_click

        self.active = start_active

        self.setup_button()

    def setup_button(self) -> None:
        """Prepare the text_wrapper attribute."""
        self.text_wrapper.button_object.on_click = self.on_button_click

    def on_button_click(self) -> None:
        """Call normal_button_on_click and select this option."""
        self.text_wrapper.button_object.call_func(self.normal_button_on_click)

        self.parent_dropdown.option_selected(self)

    def update(self) -> None:
        """
        Update the object.
        
        This should be called once per frame.
        """
        if self.active:
            self.text_wrapper.update()


class Dropdown(toggle.Toggle):
    """
    The class from which all dropdown menus inherit from.

    This class should not be used on its own.
    Instead use RectDropdown, CircleDropdown, BorderedRectDropdown or BorderedCircleDropdown.

    Inherits from toggle.Toggle.

    Attributes
    ----------
    all attributes from toggle.Toggle
    font_colour : tuple[int]
        the colour of displayed text
    font_size : int
        the size of displayed text
    font_name : str | None, optional
        the path to, or name of, the font used to display text (defaults to None)
    on_option_changed : callable, optional
        the function called once the option is changed. If it accepts 1 argument, the option is passed in; if it accepts 2 arguments, the option and the dropdown object are passed in; if no arguments are accepted, the function is just called (default ot None)
    antialias : bool, optional
        whether th text is drawn with antialias
    options : lsit[Option]
        a list of all options in the dropdown menu
    selected_option : Option
        the currently selected option object
    text_wrapper : button.TextWrapper
        the button text wrapper oboject that provides the button functionality, like on click events

    Methods
    -------
    all methods from toggle.Toggle
    create_options(option_buttons: list[object], option_names: list[str], start_active: bool)
        return a list of Option objects when given a list of their names and button objects
    on_value_changed(value: bool)
        update the active attribute of each option
    option_selected(option: Option)
        change selected option and update text
    update()
        update the dropdown object
    """

    def __init__(self, button_object: object, option_buttons: list[object], option_names: list[str], font_colour: tuple[int], font_size: int, font_name: str | None = None, on_option_changed: callable = None, initial_option: int = 0, start_active: bool = False, antialias: bool = False) -> None:
        """Construct the necessary attributes for the Dropdown object."""
        self.font_colour = font_colour
        self.font_size = font_size
        self.font_name = font_name

        self.on_option_changed = on_option_changed

        self.antialias = antialias
        
        self.options = self.create_options(option_buttons, option_names, start_active)

        self.selected_option = self.options[initial_option]

        self.text_wrapper = button.TextWrapper(button_object, self.selected_option.text_wrapper.text, font_colour, font_size, font_name, antialias)

        super().__init__(button_object, self.on_value_changed, False)

    def create_options(self, option_buttons: list[object], option_names: list[str], start_active: bool) -> list[Option]:
        """Return a list of Option objects when given a list of their names and button objects."""
        options = []
        for i, x in enumerate(option_buttons):
            text = option_names[i]

            text_wrapper = button.TextWrapper(x, text, self.font_colour, self.font_size, self.font_name, self.antialias)
            option = Option(self, text_wrapper, start_active)

            options.append(option)

        return options

    def on_value_changed(self, value: bool) -> None:
        """Update the active attribute of each option."""
        for i in self.options:
            i.active = value

    def option_selected(self, option: Option) -> None:
        """Change selected option and update text."""
        self.selected_option = option
        self.selected = False

        self.text_wrapper.update_text(option.text_wrapper.text, self.font_colour, self.font_size, self.font_name)

        utils.call_func(self.on_option_changed, option, self)

    def update(self) -> None:
        """
        Update the dropdown object.

        This should be called once per frame.
        """
        self.text_wrapper.update()

        for i in self.options:
            i.update()

        if self.selected != self.prev_selected:
            utils.call_func(self.on_value_changed, self.selected)
            self.prev_selected = self.selected


class RectDropdown(Dropdown):
    """
    A dropdown menu with rectangular options.

    Inherits from Dropdown.

    Attributes
    ----------
    all attributes from Dropdown
    button_object : button.RectButton
        the button object for the main dropdown button
    option_buttons : list[button.RectButton]
        a list of the button objects for each of the options in the dropdown menu

    Methods
    -------
    all methods from Dropdown
    create_buttons(surface: pygame.Surface, num_options: int, x: int, y: int, colour: tuple[int], width: int, height: int, y_offset: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int)
        create the button objects for each of the options in the dropdown menu
    """

    def __init__(self, surface: pygame.Surface, option_names: list[str], x: int, y: int, background_colour: tuple[int], width: int, height: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, y_offset: int = 10, on_option_changed: callable = None, initial_option: int = 0, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, start_active: bool = False, antialias: bool = False) -> None:
        """Construct the necessary attributes for the RectDropdown object."""
        button_object = button.RectButton(surface, x, y, background_colour, width, height, on_click, on_hover, on_normal, corner_radius)
        option_buttons = self.create_buttons(surface, len(option_names), x, y, background_colour, width, height, y_offset, on_click, on_hover, on_normal, corner_radius)

        super().__init__(button_object, option_buttons, option_names, font_colour, font_size, font_name, on_option_changed, initial_option, start_active, antialias)

    def create_buttons(self, surface: pygame.Surface, num_options: int, x: int, y: int, colour: tuple[int], width: int, height: int, y_offset: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int) -> list[button.RectButton]:
        """Create the button objects for each of the options in the dropdown menu."""
        buttons = []
        for i in range(1, num_options + 1):
            new_y = y + i * (height + y_offset)

            btn = button.RectButton(surface, x, new_y, colour, width, height, on_click, on_hover, on_normal, corner_radius)
            buttons.append(btn)
            
        return buttons
    

class CircleDropdown(Dropdown):
    """
    A dropdown menu with circular options.

    Inherits from Dropdown.

    Attributes
    ----------
    all attributes from Dropdown
    button_object : button.RectButton
        the button object for the main dropdown button
    option_buttons : list[button.RectButton]
        a list of the button objects for each of the options in the dropdown menu

    Methods
    -------
    all methods from Dropdown
    create_buttons(surface: pygame.Surface, num_options: int, x: int, y: int, colour: tuple[int], width: int, height: int, y_offset: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int)
        create the button objects for each of the options in the dropdown menu
    """

    def __init__(self, surface: pygame.Surface, option_names: list[str], x: int, y: int, background_colour: tuple[int], radius: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, y_offset: int = 10, on_option_changed: callable = None, initial_option: int = 0, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, start_active: bool = False, antialias: bool = False) -> None:
        """Construct the necessary attributes for the CircleDropdown object."""
        button_object = button.CircleButton(surface, x, y, background_colour, radius, on_click, on_hover, on_normal)
        option_buttons = self.create_buttons(surface, len(option_names), x, y, background_colour, radius, y_offset, on_click, on_hover, on_normal)

        super().__init__(button_object, option_buttons, option_names, font_colour, font_size, font_name, on_option_changed, initial_option, start_active, antialias)

    def create_buttons(self, surface: pygame.Surface, num_options: int, x: int, y: int, colour: tuple[int], radius: int, y_offset: int, on_click: callable, on_hover: callable, on_normal: callable) -> list[button.RectButton]:
        """Create the button objects for each of the options in the dropdown menu."""
        buttons = []
        for i in range(1, num_options + 1):
            new_y = y + i * (2 * radius + y_offset)

            btn = button.CircleButton(surface, x, new_y, colour, radius, on_click, on_hover, on_normal)
            buttons.append(btn)
            
        return buttons
    

class BorderedRectDropdown(Dropdown):
    """
    A dropdown menu with rectangular options with borders.

    Inherits from Dropdown.

    Attributes
    ----------
    all attributes from Dropdown
    button_object : button.RectButton
        the button object for the main dropdown button
    option_buttons : list[button.RectButton]
        a list of the button objects for each of the options in the dropdown menu

    Methods
    -------
    all methods from Dropdown
    create_buttons(surface: pygame.Surface, num_options: int, x: int, y: int, colour: tuple[int], width: int, height: int, y_offset: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int)
        create the button objects for each of the options in the dropdown menu
    """

    def __init__(self, surface: pygame.Surface, option_names: list[str], x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], width: int, height: int, border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, y_offset: int = 10, on_option_changed: callable = None, initial_option: int = 0, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, start_active: bool = False, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedRectDropdown object."""
        button_object = button.BorderedRectButton(surface, x, y, background_colour, border_colour, width, height, border_width, on_click, on_hover, on_normal, corner_radius)
        option_buttons = self.create_buttons(surface, len(option_names), x, y, background_colour, border_colour, width, height, border_width, y_offset, on_click, on_hover, on_normal, corner_radius)

        super().__init__(button_object, option_buttons, option_names, font_colour, font_size, font_name, on_option_changed, initial_option, start_active, antialias)

    def create_buttons(self, surface: pygame.Surface, num_options: int, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], width: int, height: int, border_width: tuple[int], y_offset: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int) -> list[button.RectButton]:
        """Create the button objects for each of the options in the dropdown menu."""
        buttons = []
        for i in range(1, num_options + 1):
            new_y = y + i * (height + y_offset)

            btn = button.BorderedRectButton(surface, x, new_y, background_colour, border_colour, width, height, border_width, on_click, on_hover, on_normal, corner_radius)
            buttons.append(btn)
            
        return buttons
    

class BorderedCircleDropdown(Dropdown):
    """
    A dropdown menu with circular options with borders.

    Inherits from Dropdown.

    Attributes
    ----------
    all attributes from Dropdown
    button_object : button.RectButton
        the button object for the main dropdown button
    option_buttons : list[button.RectButton]
        a list of the button objects for each of the options in the dropdown menu

    Methods
    -------
    all methods from Dropdown
    create_buttons(surface: pygame.Surface, num_options: int, x: int, y: int, colour: tuple[int], width: int, height: int, y_offset: int, on_click: callable, on_hover: callable, on_normal: callable, corner_radius: int)
        create the button objects for each of the options in the dropdown menu
    """

    def __init__(self, surface: pygame.Surface, option_names: list[str], x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], radius: int, border_width: int, font_colour: tuple[int], font_size: int, font_name: str | None = None, y_offset: int = 10, on_option_changed: callable = None, initial_option: int = 0, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, start_active: bool = False, antialias: bool = False) -> None:
        """Construct the necessary attributes for the BorderedCircleDropdown object."""
        button_object = button.BorderedCircleButton(surface, x, y, background_colour, border_colour, radius, border_width, on_click, on_hover, on_normal)
        option_buttons = self.create_buttons(surface, len(option_names), x, y, background_colour, border_colour, radius, border_width, y_offset, on_click, on_hover, on_normal)

        super().__init__(button_object, option_buttons, option_names, font_colour, font_size, font_name, on_option_changed, initial_option, start_active, antialias)

    def create_buttons(self, surface: pygame.Surface, num_options: int, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], radius: int, border_width: int, y_offset: int, on_click: callable, on_hover: callable, on_normal: callable) -> list[button.RectButton]:
        """Create the button objects for each of the options in the dropdown menu."""
        buttons = []
        for i in range(1, num_options + 1):
            new_y = y + i * (2 * radius + y_offset)

            btn = button.BorderedCircleButton(surface, x, new_y, background_colour, border_colour, radius, border_width, on_click, on_hover, on_normal)
            buttons.append(btn)
            
        return buttons


from pygame_ui_toolkit import pygame
from pygame_ui_toolkit import utils


class Button:
    """
    The base class that all other button classes inherit from.

    This class should not be used directly. 
    Instead use RectButton, CircleButton, PolygonButton, BorderedRectButton, BorderedCircleButton, or BorderedPolygonButton.

    Attributes
    ----------
    surface : pygame.Surface
        the surface that the button is drawn onto
    x : int
        the x position of the center of the button
    y : int
        the y position of the center of the button
    on_click : callable | None, optional
        the function that is called when the button is clicked. If the function accepts 1 argument, self is passed into it (defaults to None)
    on_hover : callable | None, optional
        the function that is called when the button is hovered over. If the function accepts 1 argument, self is passed into it (defaults to None)
    on_normal : callable | None, optional
        the function that is called when the button is neither clicked or hovered over. If the function accepts 1 argument, self is passed into it (defaults to None)
    click_once : bool, optional
        if True, the on_click and on_hover functions will be called once per click or hover event, otherwise they will be called every ferame the button is clicked or hovered (defaults to True)
    
    Methods
    -------
    update_clicked(currently_clicked: bool)
        update the clicked attribute.
    update_hovered(currently_hovered: bool)
        update the hovered attribute.
    call_func(func: callable)
        call utils.call_func(func, self).
    check_click()
        return whether the button is ciurrently being clicked.
    check_hover()
        return whether the button is being hovered over.
    update()
        draw the button and call the on_click, on_hover and on_normal functions where appropriate.
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the Button object"""
        self.surface = surface

        self.x = x
        self.y = y

        self.on_click = on_click
        self.on_hover = on_hover
        self.on_normal = on_normal

        self.click_once = click_once

        self.clicked = False
        self.hovered = False

    def update_clicked(self, currently_clicked: bool) -> bool:
        """
        Update the clicked attribute.
        
        Parameters
        ----------

        currently_clicked : bool
            whether the mouse is currently over the button and the mouse button is pressed

        Returns
        -------
        bool
            whether on_click should be called
        """
        if not self.click_once:
            self.clicked = currently_clicked
            return currently_clicked
        
        if not currently_clicked:
            self.clicked = False
            return False
        
        if self.clicked:
            return False
        else:
            self.clicked = True
            return True
        
    def update_hovered(self, currently_hovered: bool) -> bool:
        """
        Update the hovered attribute.
        
        Parameters
        ----------

        currently_hovered : bool
            whether the mouse is currently over the button

        Returns
        -------
        bool
            whether on_hover should be called
        """
        if not self.click_once:
            self.hovered = currently_hovered
            return currently_hovered
        
        if not currently_hovered:
            self.hovered = False
            return False
        
        if self.hovered:
            return False
        else:
            self.hovered = True
            return True

    get_pos = lambda self: (self.x, self.y)

    def call_func(self, func: callable) -> None:
        """Call utils.call_func(func, self)."""
        if func == None:
            return
        else:
            utils.call_func(func, self)
        
    def check_click(self) -> bool:
        """Return whether the button is ciurrently being clicked."""
        mouse_over = self.mouse_over()
        mouse_down = pygame.mouse.get_pressed(3)[0]

        click = mouse_over and mouse_down

        return self.update_clicked(click)
    
    def check_hover(self) -> bool:
        """Return whether the button is currently being hovered over."""
        mouse_over = self.mouse_over()

        return self.update_hovered(mouse_over)
    
    def update(self) -> None:
        """
        Draw the button and call the on_click, on_hover and on_normal functions where appropriate.

        This should be called once per frame.

        An exception is raised if the button object does not contain the nessesary methods.
        """
        self.draw()
        
        if not hasattr(self, "check_click") or not hasattr(self, "check_hover"):
            raise Exception("Base Button class should not be used by itself. Use another class like RectButton instead")

        if self.check_click():
            self.call_func(self.on_click)
        elif self.check_hover():
            self.call_func(self.on_hover)
        else:
            self.call_func(self.on_normal)


class RectButton(Button):
    """
    A rectangular button.

    This inherits from the Button class.

    Attributes
    ----------
    This class inherits from Button, so contains all the attributes that Button does.
    It also contains these additional attributes:
    background_colour : tuple[int]
        the colour of the button
    width : int
        the width (parallel to the x axis) of the button
    corner_radius : int, optional
        the radius of the circles on the rounded corners of the rectangle where -1 means sharp edges and not rounded corners (defaults to -1)
    
    Methods
    -------
    This class inherits from Button, so contains all the metehods that Button does.
    It also contains these additional methods:
    mouse_over()
        return whether the mouse is colliding with, or within the region of, the button rectangle
    draw()
        draw a rectangle with the appropriate width, height, position and colour
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], width: int, height: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, click_once: bool = True) -> None:
        """Construct the necessary attributes for the RectButton object"""
        super().__init__(surface, x, y, on_click, on_hover, on_normal, click_once)

        self.width = width
        self.height = height

        self.background_colour = background_colour

        self.corner_radius = corner_radius

    def mouse_over(self) -> bool:
        """Return whether the mouse is colliding with, or within the region of, the button rectangle"""
        x, y = pygame.mouse.get_pos()

        min_x = self.x - self.width // 2
        max_x = self.x + self.width // 2
        min_y = self.y - self.height // 2
        max_y = self.y + self.height // 2

        return min_x <= x <= max_x and min_y <= y <= max_y
    
    def draw(self) -> None:
        """Draw a rectangle with the appropriate width, height, position and colour"""
        rect = pygame.Rect(self.x - self.width // 2, self.y - self.height // 2, self.width, self.height)
        pygame.draw.rect(self.surface, self.background_colour, rect, border_radius=self.corner_radius)

    get_width = lambda self: self.width


class CircleButton(Button):
    """
    A circular button.

    This inherits from the Button class.

    Attributes
    ----------
    This class inherits from Button, so contains all the attributes that Button does.
    It also contains these additional attributes:
    background_colour : tuple[int]
        the colour of the button
    radius : int
        the radius of the circle
        
    Methods
    -------
    This class inherits from Button, so contains all the metehods that Button does.
    It also contains these additional methods:
    mouse_over()
        return whether the mouse position is within the circle
    draw()
        draw a circle with the appropriate radius, position and colour
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], radius: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the nessesary attributes for the CircleButton object"""
        super().__init__(surface, x, y, on_click, on_hover, on_normal, click_once)

        self.background_colour = background_colour

        self.radius = radius
        self.radius_sq = radius**2

    def mouse_over(self) -> bool:
        """Return whether the mouse position is within the circle"""
        x, y = pygame.mouse.get_pos()

        dist_sq = (x - self.x)**2 + (y - self.y)**2

        return dist_sq < self.radius_sq
    
    def draw(self) -> None:
        """Draw a circle with the appropriate radius, position and colour"""
        center = self.get_pos()
        pygame.draw.circle(self.surface, self.background_colour, center, self.radius)

    get_width = lambda self: self.radius * 2


class PolygonButton(Button):
    """
    A button in the shape of any convex polygon.
    Convex polygons are **not** supported and will result in unexpected behaviour.

    This inherits from the Button class.

    Attributes
    ----------
    This class inherits from Button, so contains all the attributes that Button does.
    It also contains these additional attributes:
    background_colour: tuple[int]
        the colour of the button
    points : list[tuple[int]]
        a list of x, y coordinates that correspond to each point on the polygon
        
    Methods
    -------
    This class inherits from Button, so contains all the metehods that Button does.
    It also contains these additional methods:
    get_center(points: list[tuple[int]])
        return the center of the polygon by averaging the position of each point
    get_inequalities(points: list[tuple[int]], center_x: int, center_y: int)
        return a list of inequalities that describe the button region
    mouse_over()
        return whether the mouse position is within the polygon using inequalities
    draw()
        draw a polygon with the appropriate points and colour
    """

    def __init__(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the PolygonButton object"""
        x, y = self.get_center(points)
        
        super().__init__(surface, x, y, on_click, on_hover, on_normal, click_once)

        self.points = points
        self.background_colour = background_colour

        self.inequalities = self.get_inequalities(points, x, y)

    def get_center(self, points: list[tuple[int]]) -> tuple[int]:
        """
        Return the center of the polygon by averaging the position of each point.

        Parameters
        ----------
        points : list[tuple[int]]
            a list of x, y coordinates that correspond to each point on the polygon

        Returns
        -------
        tuple[int]
            x, y coordinates to describe the center of the polygon
        """
        x_values = [i[0] for i in points]
        y_values = [i[1] for i in points]

        x = sum(x_values) // len(x_values)
        y = sum(y_values) // len(y_values)

        return x, y
    
    def get_inequalities(self, points: list[tuple[int]], center_x: int, center_y: int) -> list[tuple[float, float, bool]]:
        """
        Return a list of inequalities that describe the button region.

        Parameters
        ----------
        points : list[tuple[int]]
            a list of x, y coordinates that correspond to each point on the polygon
        center_x : int
            the x coordinate of the center of the polygon
        center_y : int
            the y coordinate of the center of the polygon

        Returns
        -------
        list[tuple[float, float, bool]]
            a list of linear inequalities in the form: line_gradient, y_intercept, greater_than (each line is interpreted as having the equation y=mx + c)
        """
        inequalities = []
        for i, x in enumerate(points):
            try:
                m = (x[1] - points[i - 1][1]) / (x[0] - points[i - 1][0])
            except ZeroDivisionError:
                m = 1e10

            c = x[1] - m * x[0]
            greater_than = (m * center_x + c) < center_y

            inequalities.append((m, c, greater_than))

        return inequalities

    def mouse_over(self) -> None:
        """Return whether the mouse position is within the polygon using inequalities."""
        x, y = pygame.mouse.get_pos()

        for m, c, greater_than in self.inequalities:
            intersect_y = m * x + c
            actual_greater_than = intersect_y < y

            if actual_greater_than != greater_than:
                return False
            
        return True
    
    def draw(self) -> None:
        """Draw a polygon with the appropriate points and colour."""
        pygame.draw.polygon(self.surface, self.background_colour, self.points)


class ImageButton(RectButton):
    """
    A button as an image.

    This inherits from the RectButton class.

    Attributes
    ----------
    This class inherits from RectButton, so contains all the attributes that RectButton does.
    It also contains these additional attributes:
    image_path : str
        the file path to the desired image
    image : pygame.Surface
        the image of the button in a form useable for pygame

    Methods
    -------
    This class inherits from RectButton, so contains all the methods that RectButton does.
    It also contains these additional methods:
    update_image()
        update the image attribute of the object

    The following methods are overwritten:
    draw()
        blit the image to the screen
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, width: int, height: int, image_path: str, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the ImageButton object"""
        super().__init__(surface, x, y, None, width, height, on_click, on_hover, on_normal, -1, click_once)

        self.update_image(image_path)

    def update_image(self, new_image_path: str) -> None:
        """Update the image attribute of the object."""
        image = pygame.image.load(new_image_path)
        scaled_image = pygame.transform.scale(image, (self.width, self.height))

        self.image = scaled_image

    def draw(self) -> None:
        """Blit the image to the screen."""
        image_x = self.x - self.width // 2
        image_y = self.y - self.height // 2

        self.surface.blit(self.image, (image_x, image_y))


class BorderedRectButton(RectButton):
    """
    A rectangluar button with a border.

    This inherits from the RectButton class.

    Attributes
    ----------
    This class inherits from RectButton, so contains all the attributes that RectButton does.
    It also contains these additional attributes:
    border_colour : tuple[int]
        the colour of the border
    border_width : int
        the width, or thickness, of the border

    Methods
    -------
    This class inherits from RectButton, so contains all the methods that RectButton does.
    The following methods are overwritten:
    draw()
        draw a rectangle and border
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], width: int, height: int, border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, corner_radius: int = -1, click_once: bool = True) -> None:
        """Construct the necessary attributes for the BorderedRectButton object."""
        super().__init__(surface, x, y, background_colour, width, height, on_click, on_hover, on_normal, corner_radius, click_once)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw a rectangle and border."""
        rect = pygame.Rect(self.x - self.width // 2, self.y - self.height // 2, self.width, self.height)

        pygame.draw.rect(self.surface, self.background_colour, rect, border_radius=self.corner_radius)
        pygame.draw.rect(self.surface, self.border_colour, rect, self.border_width, self.corner_radius)


class BorderedCircleButton(CircleButton):
    """
    A circular button with a border.

    This inherits from the CircleButton class.

    Attributes
    ----------
    This class inherits from CircleButton, so contains all the attributes that CircleButton does.
    It also contains these additional attributes:
    border_colour : tuple[int]
        the colour of the border
    border_width : int
        the width, or thickness, of the border

    Methods
    -------
    This class inherits from CircleButton, so contains all the methods that CircleButton does.
    The following methods are overwritten:
    draw()
        draw a circle and border
    """

    def __init__(self, surface: pygame.Surface, x: int, y: int, background_colour: tuple[int], border_colour: tuple[int], radius: int, border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the BorderedCircleButton object."""
        super().__init__(surface, x, y, background_colour, radius, on_click, on_hover, on_normal, click_once)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw a circle and border."""
        pygame.draw.circle(self.surface, self.background_colour, self.get_pos(), self.radius)
        pygame.draw.circle(self.surface, self.border_colour, self.get_pos(), self.radius, self.border_width)


class BorderedPolygonButton(PolygonButton):
    """
    A polygon button with a border.

    This inherits from the PolygonButton class.

    Attributes
    ----------
    This class inherits from PolygonButton, so contains all the attributes that PolygonButton does.
    It also contains these additional attributes:
    border_colour : tuple[int]
        the colour of the border
    border_width : int
        the width, or thickness, of the border

    Methods
    -------
    This class inherits from PolygonButton, so contains all the methods that PolygonButton does.
    The following methods are overwritten:
    draw()
        draw a polygon and border
    """
    def __init__(self, surface: pygame.Surface, points: list[tuple[int]], background_colour: tuple[int], border_colour: tuple[int], border_width: int, on_click: callable = None, on_hover: callable = None, on_normal: callable = None, click_once: bool = True) -> None:
        """Construct the necessary attributes for the BorderedPolygonButton object."""
        super().__init__(surface, points, background_colour, on_click, on_hover, on_normal, click_once)

        self.border_colour = border_colour
        self.border_width = border_width

    def draw(self) -> None:
        """Draw a polygon and border."""
        pygame.draw.polygon(self.surface, self.background_colour, self.points)
        pygame.draw.polygon(self.surface, self.border_colour, self.points, self.border_width)


class TextWrapper:
    """
    A wrapper object that allows text to be drawn on top of button objects.

    All buttons other than PolygonButton and BorderedPolygonButton are supported.

    Attributes
    ----------
    button_object : RectButton | CircleButton | BorderedCircleButton | BorderedRectButton
        the button object that text will be drawn upon
    text : str
        the text displayed
    font_colour : tuple[int]
        the colour of the displayed text
    font : pygame.font.Font
        the font object used to blit text
    font_name : str | None, optional
        the name of the font used (defaults to None)
    antialias : bool, optional
        whether the text is drawn with antialias (defaults to False)
    text_surface : pygame.Surface
        the surface used to render text
    text_rect : pygame.Rect
        the rect position that the text is drawn to

    Methods
    -------
    get_text()
        return a text surface and rect object to blit text
    blit_text()
        draw text to the screen
    update_text(new_text: str | None = None, new_font_colour: tuple[int] | None = None, new_font_size: int | None = None, new_font_name: str | None = None)
        change the text, font_colour, font attributes of the object
    update()
        update the TextWrapper object
    """

    def __init__(self, button_object: RectButton | CircleButton | BorderedCircleButton | BorderedRectButton, text: str, font_colour: tuple[int], font_size: int, font_name: str | None = None, antialias: bool = False) -> None:
        """Construct necessary attributes for TextWrapper object."""
        self.button_object = button_object
        
        self.text = text
        self.font_colour = font_colour
        self.font = pygame.font.Font(font_name, font_size)

        self.antialias = antialias

        self.text_surface, self.text_rect = self.get_text()

    def get_text(self) -> tuple[pygame.Surface, pygame.Rect]:
        """Return a text surface and rect object to blit text."""
        text_surf = self.font.render(self.text, self.antialias, self.font_colour, self.button_object.background_colour)
        text_rect = text_surf.get_rect()
        text_rect.center = self.button_object.get_pos()

        return text_surf, text_rect
    
    def blit_text(self) -> None:
        """Draw text to the screen."""
        self.button_object.surface.blit(self.text_surface, self.text_rect)

    def update_text(self, new_text: str, new_font_colour: tuple[int], new_font_size: int, new_font_name: str | None) -> None:
        """Change the text, font_colour, font attributes of the object."""
        utils.update_font_attrs(self, new_text, new_font_colour, new_font_name, new_font_size)

        self.text_surface, self.text_rect = self.get_text()

    def draw(self) -> None:
        """Draw button object and blit text on top."""
        self.button_object.update()
        self.blit_text()
        
    def update(self) -> None:
        """
        Update the text wrapper button.
        
        This should be called once per frame.
        """

        self.draw()


from pygame_ui_toolkit.elements import button
from pygame_ui_toolkit import utils
from pygame_ui_toolkit import pygame


class Slider:
    """
    The base slider class that both HorizontalSlider and VerticalSlider inherit from.

    Do not use this class on its own.
    Instead, use either HorizontalSlider or VerticalSlider.

    Attributes
    ----------
    surface : pygame.Surface
        the surface which the slider is drawn to
    length : int
        the length of the long side of the slider
    width : int
        the length of the short side of the slider
    x : int
        the x coordinate of the slider's position
    y : int
        the y coordinate of the slider's position
    min_value : float
        the minimum value the slider can be
    max_value : float
        the maximum value the slider can be
    value : float
        the current slider value
    prev_value : float
        the value of the slider before it is changed
    slider_colour : tuple[int]
        the colour of the slider
    on_value_changed : callable | None, optional
        the function that is called when the slider value is changed. If it accepts 1 argument, the slider value is passed in. If it accepts 2 arguments, the slider value and self are passed in (defaults to None)
    slider_buton : object
        an object from the button module that acts as the clickable button in the slider
    normal_button_on_click : callable | None
        the button object's on_click function before it was changed in setup_button()

    Methods
    -------
    check_start_in_range(self, min_value: float, max_value: float, value: float)
        raise an exception if the value is outside of the range
    setup_button()
        prepare the slider button object for use
    on_slider_button_click()
        call the normal_button_on_click() and update button pos to mouse pos
    update()
        draw slider to the screen, update the slider button object and call on_value_changed if necessary
    """

    def __init__(self, surface: pygame.Surface, length: int, width: int, x: int, y: int, min_value: float, max_value: float, start_value: float, slider_colour: tuple[int], on_value_changed: callable = None, slider_button: object | None = None, button_colour: tuple[int] = (255, 255, 255), button_radius: int | None = None) -> None:
        """Construct the necessary attributes for the Slider object."""
        self.surface = surface
        
        self.length = length
        self.width = width

        self.x = x
        self.y = y

        self.min_value = min_value
        self.max_value = max_value
        self.value = start_value
        self.prev_value = start_value

        self.slider_colour = slider_colour

        self.on_value_changed = on_value_changed

        if slider_button == None:
            if button_radius == None:
                button_radius = width

            self.slider_button = button.CircleButton(surface, 0, 0, button_colour, button_radius)
        else:
            self.slider_button = slider_button

        self.normal_button_on_click = self.slider_button.on_click

        self.check_start_in_range(min_value, max_value, start_value)
        self.setup_button()

    def check_start_in_range(self, min_value: float, max_value: float, value: float) -> None:
        """Raise an exception if the value is outside of the range."""
        if value < min_value:
            raise Exception("Start value less than min value")
        elif value > max_value:
            raise Exception("Start value greater than max value")

    def setup_button(self) -> None:
        """Prepare the slider button object for use."""
        self.slider_button.click_once = False
        self.slider_button.on_click = self.on_slider_button_click

        x, y = self.get_button_pos(self.value)

        self.slider_button.x = int(x)
        self.slider_button.y = int(y)

    def on_slider_button_click(self) -> None:
        """Call the normal_button_on_click() and update button pos to mouse pos."""
        self.slider_button.call_func(self.normal_button_on_click)

        x, y = pygame.mouse.get_pos()

        self.slider_button.x = x
        self.slider_button.y = y

        self.clamp_button_pos()

        self.value = self.get_value()

    def update(self) -> None:
        """
        Draw slider to the screen, update the slider button object and call on_value_changed if necessary.

        This should be called once per frame.
        """
        self.draw()
        self.slider_button.update()

        if self.value != self.prev_value:
            utils.call_func(self.on_value_changed, self.value, self)

            self.prev_value = self.value

    
class HorizontalSlider(Slider):
    """
    A horizontal slider.

    This inherits from the Slider class.

    Attributes
    ----------
    All attributes are the same as the base Slider class.

    Methods
    -------
    All methods from the base Slider class
    Additional methods:
    get_button_pos()
        get the starting position of the slider button
    clamp_button_pos()
        change the button position so that it is within the slider region
    get_value()
        calculate the slider value based on the current button position
    draw()
        draw the slider bar to the screen
    """

    def __init__(self, surface: pygame.Surface, length: int, height: int, x: int, y: int, min_value: float, max_value: float, start_value: float, slider_colour: tuple[int], on_value_changed: callable = None, slider_button: object | None = None, button_colour: tuple[int] = (255, 255, 255), button_radius: int | None = None) -> None:
        """Call __init__ from parent Slider class."""
        super().__init__(surface, length, height, x, y, min_value, max_value, start_value, slider_colour, on_value_changed, slider_button, button_colour, button_radius)

    def get_button_pos(self, value: float) -> tuple[int]:
        """Get the starting position of the slider button."""
        value_range = self.max_value - self.min_value
        proportion = (value - self.min_value) / value_range
        dist_from_left = self.length * proportion

        min_x = self.x - self.length // 2

        return min_x + dist_from_left, self.y

    def clamp_button_pos(self) -> None:
        """Change the button position so that it is within the slider region."""
        min_x = self.x - self.length // 2
        max_x = self.x + self.length // 2

        if self.slider_button.x < min_x:
            self.slider_button.x = min_x
        elif self.slider_button.x > max_x:
            self.slider_button.x = max_x

        self.slider_button.y = self.y

    def get_value(self) -> float:
        """Calculate the slider value based on the current button position."""
        dist_from_left = self.slider_button.x - (self.x - self.length // 2)
        proportion = dist_from_left / self.length
        range = self.max_value - self.min_value

        return self.min_value + proportion * range
    
    def draw(self) -> None:
        """Draw the slider bar to the screen."""
        rect = pygame.Rect(self.x - self.length // 2, self.y - self.width // 2, self.length, self.width)
        pygame.draw.rect(self.surface, self.slider_colour, rect)


class VerticalSlider(Slider):
    """
    A vertical slider.

    This inherits from the Slider class.

    Attributes
    ----------
    All attributes are the same as the base Slider class.

    Methods
    -------
    All methods from the base Slider class
    Additional methods:
    get_button_pos()
        get the starting position of the slider button
    clamp_button_pos()
        change the button position so that it is within the slider region
    get_value()
        calculate the slider value based on the current button position
    draw()
        draw the slider bar to the screen
    """

    def __init__(self, surface: pygame.Surface, length: int, width: int, x: int, y: int, min_value: float, max_value: float, start_value: float, slider_colour: tuple[int], on_value_changed: callable = None, slider_button: object | None = None, button_colour: tuple[int] = (255, 255, 255), button_radius: int | None = None) -> None:
        """Call __init__ from parent Slider class."""
        super().__init__(surface, length, width, x, y, min_value, max_value, start_value, slider_colour, on_value_changed, slider_button, button_colour, button_radius)

    def get_button_pos(self, value: float) -> tuple[int]:
        """Get the starting position of the slider button."""
        value_range = self.max_value - self.min_value
        proportion = (value - self.min_value) / value_range
        dist_from_top = self.length * proportion

        min_y = self.y - self.length // 2

        return self.x, min_y + dist_from_top

    def clamp_button_pos(self) -> None:
        """Change the button position so that it is within the slider region."""
        min_y = self.y - self.length // 2
        max_y = self.y + self.length // 2

        if self.slider_button.y < min_y:
            self.slider_button.y = min_y
        elif self.slider_button.y > max_y:
            self.slider_button.y = max_y

        self.slider_button.x = self.x

    def get_value(self) -> float:
        """Calculate the slider value based on the current button position."""
        dist_from_top = self.slider_button.y - (self.y - self.length // 2)
        proportion = dist_from_top / self.length
        range = self.max_value - self.min_value
        value = self.min_value + proportion * range

        return self.max_value - value
    
    def draw(self) -> None:
        """Draw the slider bar to the screen."""
        rect = pygame.Rect(self.x - self.width // 2, self.y - self.length // 2, self.width, self.length)
        pygame.draw.rect(self.surface, self.slider_colour, rect)